#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Defines the schema, defaults and validation operations for the data consumed and produced by the :class:`~wltp.experiment.Experiment`.

The model-instance is managed by :class:`pandel.Pandel`.

Example-code to get WLTP-data::

    from wltp.model  import _get_wltc_data
    cycle_data = _get_wltc_data()

    for cls, value in cycle_data['classes'].items():
        cycle = np.array(value['cycle'])
        print('%s: \n%s' % (cls, cycle))
"""

from __future__ import division, print_function, unicode_literals

from collections.abc import Mapping, Sized
import json
import logging
from textwrap import dedent

from jsonschema import (RefResolver, ValidationError)
import jsonschema
from numpy import ndarray
from pandas.core.generic import NDFrame
from six import string_types
from wltp.cycles import (class1, class2, class3)
from wltp.pandel import PandelVisitor

import itertools as it
import numpy as np
import operator as ops
import pandas as pd

try:
    from pandas.core.common import PandasError
except ImportError:
    ## Pandas-0.20.1 dropped this classs.
    #  See https://github.com/pydata/pandas-datareader/issues/305
    PandasError = ValueError


log = logging.getLogger(__name__)

def make_json_defaulter(pd_method):
    def defaulter(o):
        if (isinstance(o, np.ndarray)):
            s = o.tolist()
        elif (isinstance(o, NDFrame)):
            if pd_method is None:
                s = json.loads(pd.DataFrame.to_json(o))
            else:
                method = ops.methodcaller(pd_method)
                s = '%s:%s' % (type(o).__name__, method(o))
        else:
            s = repr(o)
        return s

    return defaulter

def json_dumps(obj, pd_method=None, **kwargs):
    return json.dumps(obj, default=make_json_defaulter(pd_method), **kwargs)
def json_dump(obj, fp, pd_method=None, **kwargs):
    json.dump(obj, fp, default=make_json_defaulter(pd_method), **kwargs)

def _get_model_base():
    """The base model for running a WLTC experiment.

    It contains some default values for the experiment (ie the default 'full-load-curve' for the vehicles).
    But note that it this model is not valid - you need to override its attributes.

    :return: a tree with the default values for the experiment.
    """

    ## Form Heinz-db
    petrol = [
        0.1       , 0.11100069, 0.12200138, 0.13300206, 0.14400275,
        0.15500344, 0.16600413, 0.17700482, 0.1880055 , 0.19900619,
        0.21000688, 0.22100757, 0.23200826, 0.24300895, 0.25400963,
        0.26501032, 0.27506528, 0.28512023, 0.29517519, 0.30523015,
        0.3152851 , 0.3231181 , 0.3309511 , 0.3387841 , 0.3466171 ,
        0.3544501 , 0.3622831 , 0.37011609, 0.37794909, 0.38578209,
        0.39361509, 0.40268373, 0.41175238, 0.42082102, 0.42988967,
        0.43895831, 0.44802696, 0.4570956 , 0.46616424, 0.47523289,
        0.48430153, 0.49385337, 0.5034052 , 0.51295704, 0.52250887,
        0.53206071, 0.54161255, 0.55116438, 0.56071622, 0.57026805,
        0.57981989, 0.58873474, 0.59764959, 0.60656444, 0.61547929,
        0.62439414, 0.633309  , 0.64222385, 0.6511387 , 0.66005355,
        0.6689684 , 0.67815415, 0.6873399 , 0.69652565, 0.70571139,
        0.71489714, 0.72408289, 0.73326864, 0.74245439, 0.75164013,
        0.76082588, 0.77026954, 0.77971321, 0.78915687, 0.79860054,
        0.8080442 , 0.81748786, 0.82693153, 0.83637519, 0.84581885,
        0.85526252, 0.86393898, 0.87261544, 0.8812919 , 0.88996837,
        0.89864483, 0.90732129, 0.91599775, 0.92467422, 0.93335068,
        0.94202714, 0.94782443, 0.95362171, 0.959419  , 0.96521628,
        0.97101357, 0.97681086, 0.98260814, 0.98840543, 0.99420271,
        1.        , 0.99556631, 0.99113261, 0.98669892, 0.98226523,
        0.97783154, 0.97339784, 0.96896415, 0.96453046, 0.96009677,
        0.95566307, 0.94509677, 0.93453046, 0.92396415, 0.91339784,
        0.90283154, 0.89226523, 0.88169892, 0.87113261, 0.86056631,
        0.85
    ]
    ## Form Heinz-db
    diesel = [
        0.1       , 0.11632768, 0.13265536, 0.14898304, 0.16531072,
        0.1816384 , 0.19796609, 0.21429377, 0.23062145, 0.24694913,
        0.26327681, 0.27956221, 0.29584762, 0.31213302, 0.32841843,
        0.34470383, 0.36098924, 0.37727464, 0.39356004, 0.40984545,
        0.42613085, 0.44289144, 0.45965203, 0.47641262, 0.49317321,
        0.5099338 , 0.52669439, 0.54345498, 0.56021557, 0.57697616,
        0.59373675, 0.60775731, 0.62177786, 0.63579841, 0.64981897,
        0.66383952, 0.67786007, 0.69188063, 0.70590118, 0.71992173,
        0.73394229, 0.74334518, 0.75274808, 0.76215098, 0.77155387,
        0.78095677, 0.79035967, 0.79976256, 0.80916546, 0.81856836,
        0.82797126, 0.83355082, 0.83913039, 0.84470996, 0.85028953,
        0.8558691 , 0.86144867, 0.86702823, 0.8726078 , 0.87818737,
        0.88376694, 0.88861739, 0.89346785, 0.8983183 , 0.90316876,
        0.90801921, 0.91286967, 0.91772013, 0.92257058, 0.92742104,
        0.93227149, 0.93551314, 0.93875479, 0.94199644, 0.94523809,
        0.94847974, 0.95172139, 0.95496304, 0.95820469, 0.96144634,
        0.96468798, 0.96645067, 0.96821335, 0.96997604, 0.97173872,
        0.97350141, 0.97526409, 0.97702678, 0.97878946, 0.98055215,
        0.98231483, 0.98408335, 0.98585187, 0.98762038, 0.9893889 ,
        0.99115742, 0.99292593, 0.99469445, 0.99646297, 0.99823148,
        1.        , 0.98871106, 0.97742212, 0.96613319, 0.95484425,
        0.94355531, 0.93226637, 0.92097743, 0.9096885 , 0.89839956,
        0.88711062, 0.88339956, 0.8796885 , 0.87597743, 0.87226637,
        0.86855531, 0.86484425, 0.86113319, 0.85742212, 0.85371106,
        0.85
    ]

    n_norm = np.arange(0.0, 1.21, 0.01)
#        petrol = np.polyval([-1.0411, 1.3853, -0.5647, 1.1107, 0.0967], n_norm).tolist()
#        diesel = np.polyval([-0.909, 1.9298, -2.2212, 2.088, 0.095], n_norm).tolist()
    default_load_curve = {
        'n_norm': n_norm,
        'p_norm': petrol
    }

    instance = {
        'vehicle': {
            "unladen_mass": None,
            "test_mass":    None,
            "v_max":    None,
            "p_rated":  None,
            "n_rated":  None,
            "n_idle":   None,
            "n_min":    None,
            "gear_ratios":[],
            'full_load_curve': default_load_curve, # FIXME: Decide load_curve by engine-type!
        },
        'params': {
            'resistance_coeffs_regression_curves': [
                [1.40E-01, 7.86E-01],
                [2.75E-05, -3.29E-02],
                [1.11E-05, 2.03E-02]
            ],
            'f_downscale_threshold':    0.01,
            'driver_mass':              75,         # kg
            'v_stopped_threshold':      1,          # km/h, <=
            'f_inertial':               1.1,
            'f_safety_margin':          0.9,  # TODO: this must change to 1-SM = 0.1
            'f_n_max':                  1.2,
            'f_n_min':                  0.125,
            'f_n_min_gear2':            0.9,
            'f_n_clutch_gear2':         [1.15, 0.03],

            'wltc_data':                _get_wltc_data(),
        }
    }

    return instance


def default_vehicle():
    return _get_model_base()['vehicle']

def default_load_curve():
    return default_vehicle()['full_load_curve']


def _get_wltc_data():
    """The WLTC-data required to run an experiment (the class-cycles and their attributes)..

    Prefer to access wltc-data through :samp:`{model}['wltc_data']`.

    :return: a tree
    """

    ## See schemas for explainations.
    ##
    wltc_data = {
        'classes': {
            'class1': class1.class_data(),
            'class2': class2.class_data(),
            'class3a': class3.class_data_a(),
            'class3b': class3.class_data_b(),
        },
    }

    return wltc_data


_url_model = '/model'
_url_wltc = '/wltc'
def _get_model_schema(additional_properties=False, for_prevalidation=False):
    """
    :param bool additional_properties: when False, 4rd-step(validation) will scream on any non-schema property found.
    :return: The json-schema(dict) for input/output of the WLTC experiment.
    """

    schema = {
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'id': _url_model,
        'title': 'Json-schema describing the input for a WLTC experiment.',
        'type': 'object', 'additionalProperties': additional_properties,
        'required': ['vehicle'],
        'properties': {
            'vehicle': {
                'title': 'vehicle model',
                'type': 'object', 'additionalProperties': additional_properties,
                'required': ['test_mass', 'v_max', 'p_rated', 'n_rated', 'n_idle', 'gear_ratios', 'full_load_curve'],
                'description': 'The vehicle attributes required for generating the WLTC velocity-profile downscaling and gear-shifts.',
                'properties': {
                    'id': {
                        'title': 'Any identifier for the object',
                        'type': ['integer', 'string'],
                    },
                    'unladen_mass': {
                        'title': 'vehicle unladen mass',
                        'type': ['number', 'null'],
                        'minimum': 0,
                        'exclusiveMinimum': True,
                        'description': dedent("""
                            The mass (kg) of the vehicle without the driver, used to decide its class,
                            as defined in Annex-4
                            """),
                    },
                    'test_mass': {
                        'title': 'vehicle test mass',
                        '$ref': '#/definitions/positiveNumber',
                        'description': dedent("""
                            The test mass of the vehicle used in all calculations (kg),
                            as defined in Annex 4.2.1.3.1, pg 94.
                           """),
                    },
                    'v_max': {
                        'title': 'maximum vehicle velocity',
                        'type': ['number', 'null'],
                        'minimum': 0,
                        'exclusiveMinimum': True,
                        'description': dedent("""
                            The maximum velocity as declared by the manufacturer.
                            If ommited, calculated as:

                                v_max = (n_rated * f_n_max (=1.2)) / gear_ratio[last]
                        """),
                    },
                    'p_rated': {
                        'title': 'maximum rated power',
                        '$ref': '#/definitions/positiveNumber',
                        'description': 'The maximum rated engine power (kW) as declared by the manufacturer.',
                    },
                    'n_rated': {
                        'title': 'rated engine revolutions',
                        '$ref': '#/definitions/positiveNumber',
                        'description': dedent("""
                            The rated engine revolutions at which an engine develops its maximum power.
                            If the maximum power is developed over an engine revolutions range,
                            it is determined by the mean of this range.
                            This is called 's' in the specs.
                        """),
                    },
                    'n_idle': {
                        'title': 'idling revolutions',
                        '$ref': '#/definitions/positiveNumber',
                        'description': 'The idling engine revolutions (Annex 1).',
                    },
                    'n_min': {
                        'title': 'minimum engine revolutions',
                        'type': ['array', 'integer', 'null'],
                        'description': dedent("""
                        Either a number with the minimum engine revolutions for gears > 2 when the vehicle is in motion,
                        or an array with the exact `n_min` for each gear (array must have length equal to gears).

                        If unspecified, the minimum `n` for gears > 2 is determined by the following equation:
                            n_min = n_idle + f_n_min(=0.125) * (n_rated - n_idle)
                        Higher values may be used if requested by the manufacturer, by setting this one.
                       """),
                    },
                    'gear_ratios': {
                        'title': 'gear ratios',
                        '$ref': '#/definitions/positiveNumbers',
                        'maxItems': 24,
                        'minItems': 3,
                        'description':
                        'An array with the gear-ratios obtained by dividing engine-revolutions (1/min) by vehicle-velocity (km/h).',
                    },
                    'resistance_coeffs': {
                        'title': 'driving resistance coefficients',
                        'type': ['null', 'array'],
                        'items': {'type': 'number'},
                        'minItems': 3,
                        'maxItems': 3,
                        'description': dedent("""
                            The 3 driving resistance coefficients f0, f1, f2,
                            in N, N/(km/h), and N/(km/h)² respectively (Annex 4).

                            If not specified, they are determined based on ``test_mass`` from
                            a pre-calculated regression curve:

                                f0 = a00 * test_mass + a01,
                                f1 = a10 * test_mass + a11,
                                f2 = a20 * test_mass + a21,

                            where ``a00, ..., a22`` specified in ``/params``.
                        """),
                    },
                    'full_load_curve': {
                        'title': 'full load power curve',
                        'description': dedent("""
                            An array/dict/dataframe holding the full load power curve in (at least) 2 columns
                            Example:

                                np.array([
                                    [ 0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 110, 120 ],
                                    [ 6.11, 21.97, 37.43, 51.05, 62.61, 72.49, 81.13, 88.7, 94.92, 98.99, 100., 96.28, 87.66 ]
                                ]).T

                            * The 1st column or `n_norm` is the normalized engine revolutions, within [0.0, 1.20]:

                                        n_norm = (n - n_idle) / (n_rated  - n_idle)

                            * The 2nd column or `p_norm` is the normalised values of the full-power load against the p_rated,
                              within [0, 1]: 
                              
                                        p_norm = p / p_rated
                        """),
                        'type': [ 'object', 'array', 'null'],
                    },
                    'pmr': {
                        'title': 'Power to Unladen-Mass',
                        'description': 'Power/unladen-Mass ratio (W/kg).',
                        'type': 'number',
                    },
                    'wltc_class': {
                        'description': 'The name of the WLTC-class (found within WLTC-data/classes) as selected by the experiment.',
                        'type': 'string',
                        'enum': ['class1', 'class2', 'class3a', 'class3b'],
                    },
                }  #veh-props
            }, # veh
            'params': {
                'title': 'experiment parameters',
                'type': 'object', 'additionalProperties': additional_properties,
                'required': [
                    'resistance_coeffs_regression_curves',
                    'driver_mass',
                    'v_stopped_threshold',
                    'f_inertial',
                    'f_safety_margin',
                    'f_n_max',
                    'f_n_min',
                    'f_n_min_gear2',
                    'f_n_clutch_gear2',
                    'wltc_data',
                ],
                'properties': {
                    'resistance_coeffs_regression_curves': {
                        'description': "Regression curve factors for calculating vehicle's ``resistance_coeffs`` when missing.",
                        'type': 'array',
                        'minItems': 3,
                        'maxItems': 3,
                        'items': {
                            'type': 'array',
                            'minItems': 2,
                            'maxItems': 2,
                            'items': {'type': 'number'},
                        },
                    },
                    'f_downscale_threshold': {
                        'title': "Downscale-factor threshold",
                        'description': "The limit for the calculated ``f_downscale``` below which no downscaling happens.",
                        'type': [ 'number', 'null'],
                        'default': 0.01,
                    },
                    'driver_mass': {
                        'title': "Driver's mass (kg)",
                        'description': "The mass (kg) of the vehicle's driver, where: Unladen_mass = (Test_mass - driver_mass) (Annex 1-3.2.6, p9).",
                        'type': [ 'number', 'null'],
                        'default': 75,
                    },
                    'v_stopped_threshold': {
                        'description': 'Velocity (km/h) under which (<=) to idle gear-shift (Annex 2-3.3, p71).',
                        'type': [ 'number', 'null'],
                        'default': 1,
                    },
                    'f_inertial': {
                        'description': "This is the 'kr' inertial-factor used in the 2nd part of the formula for calculating required-power (Annex 2-3.1, p71).",
                        'type': [ 'number', 'null'],
                        'default': 1.1,
                    },
                    'f_safety_margin': {
                        'description': dedent("""
                            Safety-margin factor for load-curve due to transitional effects (Annex 2-3.3, p72).
                            If array, its length must match those of the `gear_ratios`.
                        """),
                        'type': [ 'array', 'number', 'null'],
                        'default': 0.9,
                    },
                    'f_n_max': {
                        'description': 'For each gear, N :< n_max = n_idle + f_n_max * n_range',
                        'type': [ 'number', 'null'],
                        'default': 1.2,
                    },
                    'f_n_min': {
                        'description': 'For each gear > 2, N :> n_min = n_idle + f_n_min * n_range (unless ``n_min`` overriden by manufacturer)',
                        'type': [ 'number', 'null'],
                        'default': 0.125,
                    },
                    'f_n_min_gear2': {
                        'description': 'Gear-2 is invalid when N :< f_n_min_gear2 * n_idle.',
                        'type': [ 'number', 'null'],
                        'default': 0.9,
                    },
                    'f_n_clutch_gear2': {
                        'description': dedent("""
                            A 2-value number-array(f1, f2) controlling when to clutch gear-2:
                                N < n_clutch_gear2 := max(f1 * n_idle, f2 * n_range + n_idle),
                            unless "clutched"...
                        """),
                        'type': [ 'array', 'null'],
                        'default': [1.15, 0.03],
                    },
                    'f_downscale': {
                        'description': 'The downscaling-factor as calculated by the experiment (Annex 1-7.3, p68).',
                        'type': 'number',
                    },
                    'wltc_data': {'$ref': _url_wltc},
                }
            },
            'cycle_run': {}, #TODO: results(cycle) model-schema.
        },
        'definitions': {
            'positiveInteger': {
                'type': 'integer',
                'minimum': 0,
                'exclusiveMinimum': True,
            },
            'positiveNumber': {
                'type': 'number',
                'minimum': 0,
                'exclusiveMinimum': True,
            },
            'positiveIntegers': {
                'type': 'array',
               'items': { '$ref': '#/definitions/positiveInteger' },
            },
            'positiveNumbers': {
                'type': 'array',
               'items': { '$ref': '#/definitions/positiveNumber' },
            },
        }
    }

    return schema

def _get_wltc_schema():
    """The json-schema for the WLTC-data required to run a WLTC experiment.

    :return :dict:
    """

    schema = {
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'id': _url_wltc,
        'title': 'WLTC data',
        'type': 'object', 'additionalProperties': False,
        'required': ['classes'],
        'properties': {
            'classes': {
                'type': 'object', 'additionalProperties': False,
                'required': ['class1', 'class2', 'class3a', 'class3b'],
                'properties': {
                    'class1': {'$ref': '#definitions/class'},
                    'class2': {'$ref': '#definitions/class'},
                    'class3a': {'$ref': '#definitions/class'},
                    'class3b': {'$ref': '#definitions/class'},
                }
            },
        },
        'definitions': {
            'class': {
                'title': 'WLTC class data',
                'type': 'object', 'additionalProperties': False,
                'required': ['pmr_limits', 'parts', 'downscale', 'cycle'],
                'properties': {
                    'pmr_limits': {
                        'title': 'PMR (low, high]',
                        'description': 'Power_To_unladen-Mass ratio-limits ((low, high], W/kg) used to select classes (Annex 1, p19).',
                        'type': 'array',
                        'items': {'type': 'number'},
                        'minItems': 2, 'maxItems': 2,
                    },
                    'velocity_limits': {
                        'description': 'Velocity-limits ([low, high), km/h) within which (<) version-a/b from class3 is selected (Annex 1, p19).',
                        'type': 'array',
                        'items': {'type': 'number'},
                        'minItems': 2, 'maxItems': 2,
                    },
                    'parts': {
                        'type': 'array', 'items': {
                            'type': 'array',
                            'items': {'type': 'integer', },
                            'minItems': 2, 'maxItems': 2,
                        }
                    },
                    'downscale': {
                        'type': 'object', 'additionalProperties': False,
                        'required': ['phases', 'p_max_values', 'factor_coeffs'],
                        'properties': {
                            'phases': {
                                'type': 'array', 'additionalItems': False,
                                'items': {'$ref': 'model#definitions/positiveInteger'},
                                'maxItems': 3, 'minItems': 3,
                            },
                            'decel_phase': {
                                'type': 'array', 'additionalItems': False,
                                'items': {'type': 'integer'},
                                'maxItems': 2, 'minItems': 2,
                            },
                            'p_max_values': {
                                'type': 'array', 'additionalItems': False,
                                'items': { 'type': 'number'},
                                'maxItems': 3, 'minItems': 3,
                            },
                            'factor_coeffs': {
                                'type': 'array',
                            },
                            'v_max_split': {
                                'description': 'Velocity-limit (<, km/h) for calculating class-2 & 3 ``f_downscaling`` (Annex 1.7.3, p68).',
                                'type': 'number',
                            },
                        }
                    },
                    'checksum': { 'type': 'number'},
                    'cycle': {
                        'type': 'array',
                        'items': {'type': 'number', },
                        'minItems': 906, # class1's length
                    }
                },
            }
        }
    }

    return schema



def get_class_part_names(cls_name=None):
    """
    :param str cls_name: one of 'class1', ..., 'class3b', if missing, returns all 4 part-names
    """
    part_names = ['Low', 'Medium', 'High', 'ExtraHigh']

    if cls_name:
        wltc_data = _get_wltc_data()
        cls = wltc_data['classes'][cls_name]
        part_names = part_names[:len(cls['parts'])]

    return part_names

def get_class_parts_limits(cls_name, mdl=None, edges=False):
    """
    Parses the supplied in wltc_data and extracts the part-limits for the specified class-name.

    :param str cls_name: one of 'class1', ..., 'class3b'
    :param mdl: the mdl to parse wltc_data from, if ommited, parses the results of :func:`_get_wltc_data()`
    :param edges: when `True`, embeds internal limits into (0, len)
    :return: a list of ints with the part-limits, ie for class-3a these are 3 numbers
    """
    if mdl:
        wltc_data = mdl['wltc_data']
    else:
        wltc_data = _get_wltc_data()

    cls = wltc_data['classes'][cls_name]
    parts = cls['parts']
    parts = parts if edges else parts[:-1]
    part_limits = [end + 0.5 for (start, end) in parts]
    if edges:
        part_limits.insert(0, 0)

    return part_limits

def get_class_parts_index(cls_name, index=None, mdl=None):
    """
    Returns an array equally sized as `index` with zero-based ints denoting the part each second of the cycle belong to.

    :param str cls_name: one of 'class1', ..., 'class3b'
    :param list/array index: (Optional) the index to "segment" into parts, defaults to 1Hz class's index
    :return: a numpy-array of integers with length equal to `index`, or if not given,
                        the default length of the requested class otherwise

    Get class-checksums example::

        >>> from wltp import model
        >>> import pandas as pd

        >>> cls = 'class2'
        >>> part_limits = model.get_class_parts_index(cls)
        >>> part_limits
        array([   0,  589, 1022, 1477, 1800])

        >>> cls_data = model._get_wltc_data()['classes'][cls]
        >>> cycle = pd.DataFrame(cls_data['cycle'])
        >>> cycle.groupby(pd.cut(cycle.index, part_limits)).sum()
                            0
        (0, 589]      11162.2
        (589, 1022]   17054.3
        (1022, 1477]  24450.6
        (1477, 1800]  28869.8

    """
    limits = get_class_parts_limits(cls_name, mdl=mdl, edges=True)
    limits = np.array(limits).astype(np.int)

    if index:
        index = np.asarray(index)

    return limits

def get_class_pmr_limits(mdl=None, edges=False):
    """
    Parses the supplied in wltc_data and extracts the part-limits for the specified class-name.

    :param mdl: the mdl to parse wltc_data from, if omitted, parses the results of :func:`_get_wltc_data()`
    :param edges: when `True`, embeds internal limits into (0, len)
    :return: a list with the pmr-limits (2 numbers)
    """
    if mdl:
        wltc_data = mdl['wltc_data']
    else:
        wltc_data = _get_wltc_data()

    pmr_limits_pairs = [cls['pmr_limits'] for cls in wltc_data['classes'].values()]
    pmr_limits = sorted(set(it.chain(*pmr_limits_pairs)))
    if not edges:
        pmr_limits = pmr_limits[1:-1]    ## Exclude 0 and inf

    return pmr_limits


def merge(a, b, path=[]):
    """'merges b into a"""

    for key in b:
        bv = b[key]
        if key in a:
            av = a[key]
            if isinstance(av, Mapping) != isinstance(bv, Mapping):
#                 log.debug("Dict-values conflict at '%s'! a(%s) != b(%s)",
#                                 '/'.join(path + [str(key)]), type(av), type(bv))
                pass
            elif av is bv:
                continue # same leaf value
            elif isinstance(av, Mapping):
                merge(av, bv, path + [str(key)])
                continue
        a[key] = bv
    return a

# works
# print(merge({1:{"a":"A"},2:{"b":"B"}}, {2:{"c":"C"},3:{"d":"D"}}))
# # has conflict
# merge({1:{"a":"A"},2:{"b":"B"}}, {1:{"a":"A"},2:{"b":"C"}})



def model_validator(additional_properties=False, validate_wltc_data=False, validate_schema=False):
    schema = _get_model_schema(additional_properties)
    wltc_schema = _get_wltc_schema() if validate_wltc_data else {} ## Do not supply wltc schema, for speedup.
    resolver = RefResolver(_url_model, schema, store={_url_wltc: wltc_schema})
    validator = PandelVisitor(schema, resolver=resolver, skip_meta_validation=not validate_schema)

    return validator

def validate_model(mdl, additional_properties=False, iter_errors=False, validate_wltc_data=False, validate_schema=False):
    """
    :param bool iter_errors: does not fail, but returns a generator of ValidationErrors

    >>> validate_model(None)
    Traceback (most recent call last):
    jsonschema.exceptions.ValidationError: None is not of type 'object'
    ...

    >>> mdl = _get_model_base()
    >>> err_generator = validate_model(mdl, iter_errors=True)
    >>> sorted(err_generator, key=hash)
    [<ValidationError:
    ...

    >>> mdl = _get_model_base()
    >>> mdl["vehicle"].update({
    ...     "unladen_mass":1230,
    ...     "test_mass":   1300,
    ...     "v_max":   195,
    ...     "p_rated": 110.625,
    ...     "n_rated": 5450,
    ...     "n_idle":  950,
    ...     "n_min":   500,
    ...     "gear_ratios":[120.5, 75, 50, 43, 33, 28],
    ...     "resistance_coeffs":[100, 0.5, 0.04],
    ... })
    >>> err_generator = validate_model(mdl, iter_errors=True)
    >>> len(list(err_generator))
    0
    """

    validator = model_validator(additional_properties=additional_properties, validate_wltc_data=validate_wltc_data)
    validators = [
        validator.iter_errors(mdl),
        yield_load_curve_errors(mdl),
        yield_n_min_errors(mdl),
        yield_safety_margin_errors(mdl),
        yield_forced_cycle_errors(mdl, additional_properties)
    ]
    errors = it.chain(*[v for v in validators if not v is None])

    if iter_errors:
        return errors
    else:
        for error in errors:
            try:
                raise error
            except jsonschema.ValidationError as ex:
                ## Attempt to workround BUG: https://github.com/Julian/jsonschema/issues/164
                #
                if isinstance(ex.instance, NDFrame) or isinstance(ex.instance, ndarray):
                    ex.instance = '%s: %s' % (type(ex.instance), str(ex.instance))
                    ex.instance = str(ex.instance)
                raise



def yield_load_curve_errors(mdl):
    vehicle = mdl['vehicle']
    f_n_max = mdl['params']['f_n_max']
    wot = vehicle['full_load_curve']
    try:
        if not isinstance(wot, pd.DataFrame):
            wot = pd.DataFrame(wot)
        if wot.shape[0] < wot.shape[1]:
            wot = wot.T

        cols = wot.columns
        if wot.shape[1] == 1:
            if cols[0] != 'p_norm':
                log.warning("Assuming the single-column(%s) to be the `p_norm` and the index the `n_norm`.", cols[0])
                cols = ['p_norm']
                wot.columns = cols
            wot['n_norm'] = wot.index
            wot = wot[['n_norm', 'p_norm']]
        elif wot.shape[1] == 2:
            if not all(isinstance(i, string_types) for i in cols):
                wot.columns = ['n_norm', 'p_norm']

        n_norm = wot['n_norm']
        if (min(n_norm) > 0.1):
            log.warning('The full_load_curve SHOULD begin at least from 0%%, not from %f%%!' % min(n_norm))
        max_x_limit = f_n_max
        if (max(n_norm) < max_x_limit):
            log.warning('The full_load_curve SHOULD finish at least on f_n_max(%f%%), not on %f%%!' % (max_x_limit, max(n_norm)))

        p_norm = wot['p_norm']
        if (min(p_norm) < 0):
            yield ValidationError('The full_load_curve must not contain negative power(%f)!' % min(p_norm))
        if (max(p_norm) > 1):
            log.warning('The full_load_curve must not exceed 1, found %f!', max(p_norm))

        vehicle['full_load_curve'] = wot
    except (KeyError, PandasError) as ex:
        yield ValidationError('Invalid Full-load-curve, due to: %s' % ex, cause=ex)

def yield_n_min_errors(mdl):
    vehicle = mdl['vehicle']
    ngears = len(vehicle['gear_ratios'])
    n_min = vehicle.get('n_min')
    if not n_min is None:
        try:
                if isinstance(n_min, Sized):
                    if len(n_min) != ngears:
                        yield ValidationError("Length mismatch of n_min(%s) != gear_ratios(%s)!" % (len(n_min), ngears))
                else:
                    vehicle['n_min'] = [n_min] * ngears
        except PandasError as ex:
            yield ValidationError("Invalid 'n_min', due to: %s" % ex, cause=ex)

def yield_safety_margin_errors(mdl):
    params = mdl['params']
    f_safety_margin = params['f_safety_margin']
    ngears = len(mdl['vehicle']['gear_ratios'])
    try:
        if isinstance(f_safety_margin, Sized):
            if len(f_safety_margin) != ngears:
                yield ValidationError("Length mismatch of f_safety_margin(%s) != gear_ratios(%s)!" % (len(f_safety_margin), ngears))
        else:
            params['f_safety_margin'] = [f_safety_margin] * ngears
    except PandasError as ex:
        yield ValidationError("Invalid 'gear_n_min', due to: %s" % ex, cause=ex)


def yield_forced_cycle_errors(mdl, additional_properties):
    params = mdl['params']
    forced_cycle = params.get('forced_cycle')
    if not forced_cycle is None:
        try:
            if not isinstance(forced_cycle, pd.DataFrame):
                forced_cycle = pd.DataFrame(forced_cycle)
                if forced_cycle.shape[0] == forced_cycle.shape[1]:
                    yield ValidationError('The full_load_curve is a square matrix(%s), cannot decide orientation!' % (forced_cycle.shape,))
            if forced_cycle.shape[0] < forced_cycle.shape[1]:
                forced_cycle = forced_cycle.T
            cols = forced_cycle.columns

            # if not additional_properties and not set(cols) <= set(['v','slide']):
            #     yield ValidationError('Unexpected columns!')

            if forced_cycle.shape[1] == 1:
                if cols[0] == 1:
                    log.warning("Assuming the unamed single-column to be the velocity_profile(%s).", cols[0])
                    forced_cycle.columns = ['v']

            params['forced_cycle'] = forced_cycle
        except PandasError as ex:
            yield ValidationError('Invalid forced_cycle, due to: %s' % ex, cause=ex)

get_model_schema = _get_model_schema

if __name__ == '__main__':
    print('Model: %s' % json.dumps([_get_model_schema(), _get_wltc_schema()], indent=1))
    print('Model: %s' % json.dumps(_get_model_base(), indent=1))




#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''Build a python array from reading WLTC-data files and print it to be included as code.

It is used for preparing the WLTC variables for the default Model-data.
For example, to print (and optionally validate) the cycle for WLTC-Class 3.2, execute::

    python3 buildwltcclass.py class3.2 True

Or for printing all classes::

    python3 buildwltcclass.py

Then copy-paste the output of the program as a python-variable within the respective
 wltp.classes.classX.py modules.
'''

from __future__ import division, print_function, unicode_literals

import collections
import csv
import os, io


wltc_data_files = {
    'class1': [
        'class1-1-Low',
        'class1-2-Medium',
    ],
    'class2': [
        'class2-1-Low',
        'class2-2-Medium',
        'class2-3-High',
        'class2-4-ExtraHigh'
    ],
    'class3a': [
        'class3-1-Low',
        'class3-2a-Medium',
        'class3-3a-High',
        'class3-4-ExtraHigh'
    ],
    'class3b': [
        'class3-1-Low',
        'class3-2b-Medium',
        'class3-3b-High',
        'class3-4-ExtraHigh'
    ]
}


def read_wltc_class(wltc_class, assert_files = False):
    '''
    :argument: string: wltc_class: ( 'class1' | 'class2' | 'class3a| 'class3b' )
    :return :generator: with the speed_column floats
    '''

    def iterate_csv(fnames, yieldfun, beginfilefunc = None, endfilefunc = None):
        '''
        :argument: stringarray: fnames: the names of the resource-files
        :argument: function: yieldfunc: how to yield from each csv-row
        :return :generator with the results of the yieldfun(row)
        '''

        for fname in fnames:
            with io.open(fname, 'rb') as csvfile:  # @UndefinedVariable
                first_line = True
                reader = csv.reader(io.TextIOWrapper(csvfile))
                for row in reader:
                    if (beginfilefunc is not None and first_line):
                        beginfilefunc(row)
                    try:
                        yield yieldfun(row)
                    except csv.Error as e:
                        print('file {}, line {}: {}'.format(fname, reader.line_num, e))
                        raise
                    first_line = False

                if (endfilefunc is not None):
                    endfilefunc(row)



    if (wltc_class not in wltc_data_files):
        raise ValueError('Unknown WLTC-class(%s)!  One of (%s) expected.' %
                         (wltc_class, ', '.join(sorted(wltc_data_files.keys()))))

    mydir = os.path.dirname(__file__)
    class_part_names = [os.path.join(mydir, 'data', '%s.csv' % pname)
                        for pname in wltc_data_files[wltc_class]]

    if (assert_files):
        timer = -1;
        def assert_time_column(row):
            nonlocal timer
            timer += 1
            if (float(row[0]) != timer):
                raise AssertionError('index-col(%s) != expected(%s)' % (row[0], timer))
        def consume(iterator):
            collections.deque(iterator, maxlen=0)

        consume(iterate_csv(class_part_names, assert_time_column))


    def extract_speed_column(row):
        return float(row[1])


    def extract_start_time(row):
        nonlocal parts
        time = int(float(row[0]))
        parts.append([time])


    def extract_end_time(row):
        nonlocal parts
        time = int(float(row[0]))
        parts[-1].append(time)

    parts = []
    speed_profile = iterate_csv(class_part_names, extract_speed_column, extract_start_time, extract_end_time)

    return (speed_profile, parts)


def print_class_data(wltc_class, assert_files = False):
    '''Prepares and pretty-prints python variables with the data for the spec'ed WLTC-class.'''

    def print_speed_list(timelist):
        print('[', end='\n    ')
        i = 0
        for v in timelist:
            i += 1
            print('%.1f, ' % v, end='\n    ' if (i % 16 == 0) else '')
        print('\n]\n')

    (speed_profile, parts) = read_wltc_class(wltc_class, assert_files)
    speed_profile = list(speed_profile)

    print('\n\n>>> CLASS: %s\nvalues#  : %s\n' % (wltc_class, len(speed_profile)))
    print("'parts': {}, \n'cycle': ".format(parts))
    print_speed_list(speed_profile)




if __name__ == '__main__':
    #print('values#: %s' % list(read_wltc_class('class3.2', assert_files=True)))

    import sys

    if (len(sys.argv) > 1):
        print_class_data(*sys.argv[1:])
    else:
        for wltc_class in sorted(wltc_data_files):
            print_class_data(wltc_class, True)




#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""Preproc script that reads the tables from the UN-document as CSV, reshapes and stores them as 2-columns CSV.

This utility preprocess the WLTC-cycle speed-profile tables of the spec's word-document
into a format suitable to be consumed by wltp.
An intermediate human-UI step is required involving a spreadsheet where the word-tables
are copied into a spreadsheet to be exported as CSV-files.

It expects odd columns to contain the time-steps, so that by sorting on it to reconstruct the whole cycle-profile.

Example::
    cat multiColumn.txt | python3 csvcolumns8to2.py > out.csv
or::
    python3 csvcolumns8to2.py multiColumn.txt
"""


from __future__ import division, print_function, unicode_literals

def csvcolumns8to2(inp, inpfname, out, outfname):
    import pandas as pd

    sys.stderr.write('Transforming to 2-columns: inp(%s) --> out(%s)\n' % (inpfname, outfname))

    xls     = pd.ExcelFile(inp)
    sheets  = xls.sheet_names
    df      = xls.parse(sheets[0])
    df      = df.convert_objects(convert_numeric=True)


    ## Use numpy to split and join by 2-columns.
    #
    twocols = df.values.reshape(-1, 2)

    ## Revert back to pandas to filter empties and sort.
    #
    df = pd.DataFrame(twocols)
    df = df[pd.notnull(df[0])]
    df = df.sort(0, axis=0)

    df[[0]] = df[[0]].astype(int) # Time

    df.to_csv(out, header=False, index=False, encoding='latin')




if (__name__ == '__main__'):
    import sys
    import os

    inp = sys.stdin
    inpfname = '<stdin>'
    out = sys.stdout
    outfname = '<stdout>'
    if (len(sys.argv) >= 2):
        inpfname = inp = os.path.abspath(sys.argv[1])
    if (len(sys.argv) >= 3):
        outfname = out = os.path.abspath(sys.argv[2])

    csvcolumns8to2(inp, inpfname, out, outfname)




#!/usr/bin/env python
#-*- coding: utf-8 -*-
import os
import logging
import glob
from textwrap import dedent
from win32com.universal import com_error
import xlwings as xw


__version__ = '0.0.1'

log = logging.getLogger(__name__)
log.trace = lambda *args, **kws: log.log(0, *args, **kws)

DEFAULT_LOG_LEVEL = logging.DEBUG
def _init_logging(loglevel=DEFAULT_LOG_LEVEL):
    logging.basicConfig(level=loglevel)
    logging.getLogger().setLevel(level=loglevel)


## Probably Windows-only:
#    TODO: Should become a class-method on xw.Workbook
#    when/if it works reliably, see github/xlwings #30.
def get_active_workbook():
    from win32com.client import dynamic

    com_app = dynamic.Dispatch('Excel.Application')
    com_wb = com_app.ActiveWorkbook
    wb = xw.Workbook(xl_workbook=com_wb)

    return wb


def get_Workbook(wrkb_fname):
    """
    :param str wrkb_fname: if missing return the active excel-workbook
    """

    if wrkb_fname:
        wb = xw.Workbook(wrkb_fname)
    else:
        wb = get_active_workbook()
    return wb

def _get_xl_vb_project(xl_wb):
    """
    To allow updating Excel's VBA, follow instructions at:
      https://social.msdn.microsoft.com/Forums/office/en-US/c175d0d7-7341-4b31-9699-712e5449cf78/access-to-visual-basic-project-is-not-trusted

      1. Go to Excel's Options.
      2. Click Trust Center.
      3. Click Trust Center Settings.
      4. Click Macro Settings.
      5. Click to select the `Trust access to the VBA project object model` check box.
      6. Click OK to close the Excel Options dialog box.

    or use the Excel-2010 shortcuts
      [ALT] + [t][m][s]
    and then click the above check box.
    """
    def show_unlock_msg(msg):
        text = dedent(_get_xl_vb_project.__doc__)
        try:
            from tkinter import messagebox as msgbox
        except ImportError:
            import tkMessageBox as msgbox
        msg = '%s\n%s' % (msg, text)
        msgbox.showinfo(title="Excel Permission Denied!", message=msg)

    try:
        xl_vbp = xl_wb.VBProject
        if xl_vbp.Protection == 1:
            msg = "The VBA-code in Excel is protected!"
            msg = show_unlock_msg(msg)
            raise Exception(msg)
        return xl_vbp
    except com_error as ex: #@UndefinedVariable
        if ex.hresult == -2147352567:
            msg = "Excel complained: \n  %s" % ex.excepinfo[2]
            msg = show_unlock_msg(msg)
            raise Exception(msg)
        raise


def _gather_files(fpaths_wildcard):

    """
    :param str fpaths_wildcard: 'some/foo*bar.py'
    :return: a map {ext-less_basename --> full_path}.
    """
    def basename(fname):
        b, _ = os.path.splitext(os.path.basename(fname))
        return b

    return {basename(f): f for f in glob.glob(fpaths_wildcard)}


XL_TYPE_MODULE = 1
XL_TYPE_SHEET = 100
def _remove_vba_modules(xl_vbcs, *mod_names_to_del):
    """
    :param VBProject.VBComponents xl_vbcs:
    :param str-list mod_names_to_del: which modules to remove, assumed *all* if missing
    """

    ## Comparisons are case-insensitive.
    if mod_names_to_del:
        mod_names_to_del = map(str.lower, mod_names_to_del)

    ## Gather all workbook vba-modules.
    xl_mods = [xl_vbc for xl_vbc in xl_vbcs if xl_vbc.Type == XL_TYPE_MODULE]

    ## Remove those matching.
    #
    for m in xl_mods:
        if not mod_names_to_del or m.Name.lower() in mod_names_to_del:
            log.debug('Removing vba_module(%s)...', m.Name)
            xl_vbcs.Remove(m)


def _import_vba_files(xl_vbcs, vba_file_map):
    """
    :param VBProject.VBComponents xl_vbcs:
    :param dict vba_file_map: a map {module_name --> full_path}.

    """
    cwd = os.getcwd()
    for vba_modname, vba_fpath in vba_file_map.items():
        try:
            log.debug('Removing vba_module(%s)...', vba_modname)
            old_xl_mod = xl_vbcs.Item(vba_modname)
            xl_vbcs.Remove(old_xl_mod)
            log.info('Removed vba_module(%s).', vba_modname)
        except com_error as ex:
            log.debug('Probably vba_module(%s) did not exist, because: \n  %s', vba_modname, ex)
        log.debug('Importing vba_module(%s) from file(%s)...', vba_modname, vba_fpath)
        xl_vbc = xl_vbcs.Import(os.path.join(cwd, vba_fpath))
        log.info('Imported %i LoC for vba_module(%s) from file(%s).',
            xl_vbc.CodeModule.CountOfLines, vba_modname, vba_fpath)
        xl_vbc.Name = vba_modname


def _save_workbook(xl_workbook, path):
    ## TODO: Remove when xlwings updated to latest.
    xlOpenXMLWorkbookMacroEnabled = 52 ## From http://stackoverflow.com/questions/21306275/pywin32-saving-as-xlsm-file-instead-of-xlsx

    saved_path = xl_workbook.Path
    if (saved_path != '') and (path is None):
        # Previously saved: Save under existing name
        xl_workbook.Save()
    elif (saved_path == '') and (path is None):
        # Previously unsaved: Save under current name in current working directory
        path = os.path.join(os.getcwd(), xl_workbook.Name)
        xl_workbook.Application.DisplayAlerts = False
        xl_workbook.SaveAs(path, FileFormat=xlOpenXMLWorkbookMacroEnabled)
        xl_workbook.Application.DisplayAlerts = True
    elif path:
        # Save under new name/location
        xl_workbook.Application.DisplayAlerts = False
        xl_workbook.SaveAs(path, FileFormat=xlOpenXMLWorkbookMacroEnabled)
        xl_workbook.Application.DisplayAlerts = True


def _save_excel_as_macro_enabled(xl_wb, new_fname=None):
    DEFAULT_XLS_MACRO_FORMAT = '.xlsm'

    if not new_fname:
        _, e = os.path.splitext(xl_wb.FullName)
        if e.lower()[-1:] != 'm':
            new_fname = xl_wb.FullName + DEFAULT_XLS_MACRO_FORMAT
            log.info('Cloning as MACRO-enabled the Input-workbook(%s) --> Output-workbook(%s)',
                xl_wb.FullName, new_fname)
    _save_workbook(xl_wb, new_fname)

    return new_fname

def import_files_into_excel_workbook(infiles_wildcard, wrkb_fname=None, new_fname=None):
    """
    Add or update *xmlwings* VBA-code of a Workbook.

    :param str infiles_wildcard: 'some/foo*bar.vba'
    :param str wrkb_fname: filepath to update, active-workbook assumed if missing
    :param str new_fname: a macro-enabled (`.xlsm`, `.xltm`) filepath for updated workbook,
                            `.xlsm` gets appended if current not a macro-enabled excel-format
    """

    wb = get_Workbook(wrkb_fname=None)
    xl_wb = wb.get_xl_workbook(wb)
    xl_vbp = _get_xl_vb_project(xl_wb)
    xl_vbcs = xl_vbp.VBComponents

    infiles_map = _gather_files(infiles_wildcard)
    log.info('Modules to import into Workbook(%s): %s', wb, infiles_map)

    if infiles_map:
        #TODO: _remove_vba_modules(xl_vbcs)

        #if xl_wb.FullName.lower()[-1:] != 'm':
        #    new_fname = _save_excel_as_macro_enabled(xl_wb, new_fname=new_fname)

        _import_vba_files(xl_vbcs, infiles_map)

        _save_excel_as_macro_enabled(xl_wb, new_fname=new_fname)

    return wb

def main(*argv):
    """
    Usage::

        {cmd}  *.vbas                       # Add modules into active excel-workbook.
        {cmd}  *.vbas  in_file.xlsm         # Specify input workbook to add/update.
        {cmd}  *.vbas  foo.xlsx  bar.xlsm   # Specify input & output workbooks.
        {cmd}  *.vbas  foo.xls              # A `foo.xls.xlsm` will be created.
        {cmd}  *.vbas  foo.xls   bar        # A `bar.xlm` will be created.
        {cmd}  --pandalon inp.xls [out]     # Add 'pandalon-xlwings' modules.
    """

    _init_logging()
    cmd = os.path.basename(argv[0])
    if len(argv) < 2:
        exit('Too few arguments! \n%s' % dedent(main.__doc__.format(cmd=cmd)))
    else:
        if argv[1] == '--pandalon':
            mydir = os.path.dirname(__file__)
            argv = list(argv)
            argv[1] = os.path.join(mydir, '*.vba')
        elif argv[1].startswith('--'):
            exit('Unknown option(%s)! \n%s' % (argv[1], dedent(main.__doc__.format(cmd=cmd))))

        import_files_into_excel_workbook(*argv[1:])


if __name__ == '__main__':
    import sys
    main(*sys.argv)




#!/usr/bin/env python# -*- coding: utf-8 -*-
#
# Copyright 2013\-2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""'Setuptools script for *wltp*, the WLTC gear-shift calculator.


Install:
========
Tested with Python 3.4.

To install it, assuming you have download the sources,
do the usual::

    python setup.py install

Or get it directly from the PIP repository::

    pip install wltp
"""
# Got ideas for project-setup from many places, among others:
#    http://www.jeffknupp.com/blog/2013/08/16/open-sourcing-a-python-project-the-right-way/
#    http://python-packaging-user-guide.readthedocs.org/en/latest/current.html

import os, sys, io, re
from setuptools import setup, find_packages


# Fail early in ancient python-versions
#
py_ver = sys.version_info
if py_ver < (2, 7):
    exit("Sorry, Python2 >= 2.7 is supported!")
if py_ver >= (3,) and py_ver < (3, 3):
    exit("Sorry, Python3 >= 3.3 is supported!")
if sys.argv[-1] == "setup.py":
    exit("To install, run `python setup.py install`")

proj_name = "wltp"
mydir = os.path.dirname(__file__)


# Version-trick to have version-info in a single place,
# taken from: http://stackoverflow.com/questions/2058802/how-can-i-get-the-version-defined-in-setup-py-setuptools-in-my-package
##
def read_project_version():
    fglobals = {}
    with io.open(os.path.join(mydir, proj_name, "_version.py")) as fd:
        exec(fd.read(), fglobals)  # To read __version__
    return fglobals["__version__"]


def read_text_lines(fname):
    with io.open(os.path.join(mydir, fname)) as fd:
        return fd.readlines()


def yield_rst_only_markup(lines):
    """
    :param file_inp:     a `filename` or ``sys.stdin``?
    :param file_out:     a `filename` or ``sys.stdout`?`

    """
    substs = [
        # Selected Sphinx-only Roles.
        #
        (r":abbr:`([^`]+)`", r"\1"),
        (r":envvar:`([^`]+)`", r"``env[$\1]``"),
        (r":ref:`([^`]+)`", r"ref: *\1*"),
        (r":term:`([^`]+)`", r"**\1**"),
        (r":dfn:`([^`]+)`", r"**\1**"),
        (r":(samp|guilabel|menuselection|doc|file):`([^`]+)`", r"\1(`\2`)"),
        # Sphinx-only roles:
        #        :foo:`bar`   --> foo(``bar``)
        #        :a:foo:`bar` XXX afoo(``bar``)
        #
        # (r'(:(\w+))?:(\w+):`([^`]*)`', r'\2\3(``\4``)'),
        (r":(\w+):`([^`]*)`", r"\1(`\2`)"),
        # emphasis
        # literal
        # code
        # math
        # pep-reference
        # rfc-reference
        # strong
        # subscript, sub
        # superscript, sup
        # title-reference
        # Sphinx-only Directives.
        #
        (r"\.\. doctest", r"code-block"),
        (r"\.\. module", r"code-block"),
        (r"\.\. currentmodule::", r"currentmodule:"),
        (r"\.\. plot::", r".. plot:"),
        (r"\.\. seealso", r"info"),
        (r"\.\. glossary", r"rubric"),
        (r"\.\. figure::", r".. "),
        (r"\.\. image::", r".. "),
        (r"\.\. dispatcher", r"code-block"),
        # Other
        #
        (r"\|version\|", r"x.x.x"),
        (r"\|today\|", r"x.x.x"),
        (r"\.\. include:: AUTHORS", r"see: AUTHORS"),
    ]

    regex_subs = [(re.compile(regex, re.IGNORECASE), sub) for (regex, sub) in substs]

    def clean_line(line):
        try:
            for (regex, sub) in regex_subs:
                line = regex.sub(sub, line)
        except Exception as ex:
            print("ERROR: %s, (line(%s)" % (regex, sub))
            raise ex

        return line

    for line in lines:
        yield clean_line(line)


proj_ver = read_project_version()


readme_lines = read_text_lines("README.rst")
description = readme_lines[1].strip()
long_desc = "".join(yield_rst_only_markup(readme_lines))
# Trick from: http://peterdowns.com/posts/first-time-with-pypi.html
download_url = "https://github.com/ankostis/%s/tarball/v%s" % (proj_name, proj_ver)

plot_reqs = ["matplotlib"]
excel_reqs = ["xlwings; sys_platform == 'win32'"]
test_reqs = ["nose", "coverage", "matplotlib", "coveralls", "docopt"]
doc_reqs = ["sphinx>=1.2", "matplotlib"]  # for comparisons
dev_reqs = (
    test_reqs
    + doc_reqs
    + plot_reqs
    + excel_reqs
    + [
        "wheel",
        "twine",
        "pylint",
        # for VSCode autoformatting
        "black",
        # for VSCode RST linting
        "doc8",
        "sphinx-autobuild",
    ]
)

setup(
    name=proj_name,
    version=proj_ver,
    description=description,
    long_description=long_desc,
    long_description_content_type="text/x-rst",
    author="JRC-SRU on behalf of UNECE's GearShifting TF",
    author_email="ankostis@gmail.com",
    url="https://github.com/JRCSTU/wltp",
    download_url=download_url,
    project_urls={
        "Documentation": "https://wltp.readthedocs.io/",
        "Sources": "https://github.com/JRCSTU/wltp",
        "Bug Tracker": "https://github.com/JRCSTU/wltp/issues",
    },
    license="European Union Public Licence 1.1 or later (EUPL 1.1+)",
    keywords=[
        "automotive",
        "vehicles",
        "Light-Duty-Vehicles",
        "cars",
        "gears",
        "gearshifs",
        "simulator",
        "driving",
        "engine",
        "WLTP",
        "wltc",
        "UNECE",
        "standards",
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Development Status :: 3 - Alpha",
        "Natural Language :: English",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Manufacturing",
        "License :: OSI Approved :: European Union Public Licence 1.1 (EUPL 1.1)",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    packages=find_packages(
        exclude=["wltp.test", "wltp.test.*"]
    ),  # Wy need prune then??
    include_package_data=True,
    #     package_data= {'proj_name': ['data/*.csv']},
    #    package_data = {
    #        'wltp.excel': ['*.xlsm', '*.ico'],
    #    },
    install_requires=[
        "six",
        "jsonschema >=2.5, <3",  # 3+ dropped `validator._types`
        "numpy",
        "pandas",  # 'openpyxl', 'xlrd',
        'mock; python_version == "2.7"',
    ],
    setup_requires=[
        "setuptools-git >= 0.3"  # Gather package-data from all files in git.
    ],
    tests_require=test_reqs,
    extras_require={"plot": plot_reqs, "excel": excel_reqs, "dev": dev_reqs},
    test_suite="nose.collector",
    entry_points={"console_scripts": ["wltp = wltp.__main__:main"]},
    zip_safe=True,
    options={
        "build_sphinx": {"build_dir": "docs/_build"},
        "bdist_wheel": {"universal": True},
    },
)



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''The core that accepts a vehicle-model and wltc-classes, runs the simulation and updates the model with results (downscaled velocity & gears-profile).

.. Attention:: The documentation of this core module has several issues and needs work.

Notation
--------
* ALL_CAPITAL variables denote *vectors* over the velocity-profile (the cycle),
* ALL_CAPITAL starting with underscore (`_`) denote *matrices* (gears x time).

For instance, GEARS is like that::

    [0, 0, 1, 1, 1, 2, 2, ... 1, 0, 0]
     <----   cycle time-steps   ---->

and _GEARS is like that::

     t:||: 0  1  2  3
    ---+-------------
    g1:|[[ 1, 1, 1, 1, ... 1, 1
    g2:|   2, 2, 2, 2, ... 2, 2
    g3:|   3, 3, 3, 3, ... 3, 3
    g4:|   4, 4, 4, 4, ... 4, 4 ]]


Major vectors & matrices
------------------------

V:        floats (#cycle_steps)
    The wltp-class velocity profile.

_GEARS:    integers (#gears X #cycle_steps)
    One row for each gear (starting with 1 to #gears).

_N_GEARS:  floats (#gears X #cycle_steps)
    One row per gear with the Engine-revolutions required to follow the V-profile (unfeasable revs included),
    produced by multiplying ``V * gear-rations``.

_GEARS_YES:  boolean (#gears X #cycle_steps)
    One row per gear having ``True`` wherever gear is possible for each step.

.. Seealso:: :mod:`model` for in/out schemas
'''

from __future__ import division, unicode_literals

import logging
import re
import sys

import numpy as np
import pandas as pd

from . import model


log = logging.getLogger(__name__)


def _shapes(*arrays):
    import operator
    op_shape = operator.attrgetter('shape')
    return list(map(op_shape, arrays))

def _dtypes(*arrays):
    import operator
    op_shape = operator.attrgetter('dtype')
    return list(map(op_shape, arrays))


class Experiment(object):
    '''Runs the vehicle and cycle data describing a WLTC experiment.

    See :mod:`wltp.experiment` for documentation.
    '''


    def __init__(self, model, skip_model_validation=False, validate_wltc_data=False):
        """
        :param model:                 trees (formed by dicts & lists) holding the experiment data.
        :param skip_model_validation:  when true, does not validate the model.
        """

        self.dtype = np.float64
        self._set_model(model, skip_validation=skip_model_validation, validate_wltc_data=validate_wltc_data)

        self.wltc = self._model['params']['wltc_data']




    def run(self):
        '''Invokes the main-calculations and extracts/update Model values!

        @see: Annex 2, p 70
        '''

        model       = self._model
        vehicle     = model['vehicle']
        params      = model['params']

        ## Prepare results
        #
        results             = model.get('cycle_run')
        if results is None:
            results             = pd.DataFrame()
        else:
            results             = pd.DataFrame(results)
            log.info('Found forced `cycle-run` table(%ix%i).', results.shape[0], results.shape[1])
        model['cycle_run']  = results

        ## Ensure Time-steps start from 0 (not 1!).
        #
        results.reset_index()
        results.index.name='t'


        ## Extract vehicle attributes from model.
        #
        test_mass           = vehicle['test_mass']
        unladen_mass        = vehicle.get('unladen_mass') or test_mass - params['driver_mass']
        p_rated             = vehicle['p_rated']
        n_rated             = vehicle['n_rated']
        n_idle              = vehicle['n_idle']
        n_min_drive         = vehicle['n_min']
        v_max               = vehicle['v_max']
        gear_ratios         = vehicle['gear_ratios']
        res_coeffs          = vehicle.get('resistance_coeffs')
        if res_coeffs:
            (f0, f1, f2)        = res_coeffs
        else:
            (f0, f1, f2)        = calc_default_resistance_coeffs(test_mass, params['resistance_coeffs_regression_curves'])


        if (v_max is None):
            v_max = n_rated / gear_ratios[-1]


        is_velocity_forced      = any(col in results for col in ('v_class', 'v_target'))
        if (is_velocity_forced):
            forced_v_column         = 'v_class' if 'v_class' in results else 'v_target'
            log.info("Found forced velocity(%s).", forced_v_column)

            V                       = results[forced_v_column].values
            SLOPE                   = results.get('slope')
            if not SLOPE is None:
                SLOPE = np.asarray(SLOPE)

            results['v_class']      = V
            results['v_target']     = V
        else:
            ## Decide WLTC-class.
            #
            wltc_class = vehicle.get('wltc_class')
            if wltc_class is None:
                p_m_ratio               = (1000 * p_rated / unladen_mass)
                vehicle['pmr']          = p_m_ratio

                wltc_class              = decideClass(self.wltc, p_m_ratio, v_max)
                vehicle['wltc_class']   = wltc_class
            else:
                log.info('Found forced wltc_class(%s).', wltc_class)

            class_data          = self.wltc['classes'][wltc_class]
            V                   = np.asarray(class_data['cycle'], dtype=self.dtype)
            SLOPE               = None

            results['v_class']  = V


        ## NOTE: Improved Acceleration calc on central-values with gradient.
        #    The pure_load 2nd-part of the P_REQ from start-to-stop is 0, as it should.
        #
        #A       = np.gradient(V) ## TODO: Enable gradient acceleration-calculation.
        A       = np.diff(V); A = np.append(A, 0) # Restore element lost by diff().
        A       = A / 3.6

        ## Required-Power needed early-on by Downscaling.
        #
        f_inertial          = params.get('f_inertial', 1.1)
        P_REQ               = calcPower_required(V, A, SLOPE, test_mass, f0, f1, f2, f_inertial)

        if (not is_velocity_forced):
            ## Downscale velocity-profile.
            #
            f_downscale = params.get('f_downscale')
            if not f_downscale:
                f_downscale_threshold = params.get('f_downscale_threshold', 0.01)
                dsc_data            = class_data['downscale']
                phases              = dsc_data['phases']
                p_max_values        = dsc_data['p_max_values']
                downsc_coeffs       = dsc_data['factor_coeffs']
                dsc_v_split         = dsc_data.get('v_max_split', None)
                f_downscale         = calcDownscaleFactor(P_REQ,
                                                              p_max_values, downsc_coeffs, dsc_v_split,
                                                              p_rated, v_max,
                                                              f_downscale_threshold
                                                              )
                params['f_downscale'] = f_downscale

            if (f_downscale > 0):
                V               = downscaleCycle(V, f_downscale, phases)
            results['v_target'] = V


        ## Run cycle to find internal matrices for all gears
        #    and (optionally) gearshifts.
        #
        load_curve                  = vehicle['full_load_curve']
        (GEARS_ORIG, CLUTCH,
            _GEAR_RATIOS, _N_GEARS,
            _P_AVAILS, _N_NORMS,
            driveability_issues)    = run_cycle(V, A, P_REQ, gear_ratios,
                                               n_idle, n_min_drive, n_rated,
                                               p_rated, load_curve, params)
        results['clutch']       = CLUTCH    # TODO: Allow overridde clutch, etc.
        if ('gears_orig' in results):
            forced_gears = results['gears_orig'].values
            log.info('Found forced gears(x%i).', forced_gears.size)
            if (GEARS_ORIG.max() != len(gear_ratios)):
                raise ValueError('Forced gears(%s) specify gears(%i) > num_of_gears(%i)'%(forced_gears.shape, GEARS_ORIG.max(), len(gear_ratios)))
            GEARS_ORIG = forced_gears
        else:
            results['gears_orig']   = GEARS_ORIG

        ## Apply Driveability-rules.
        #
        GEARS                       = GEARS_ORIG.copy()
        applyDriveabilityRules(V, A, GEARS, CLUTCH, driveability_issues)

        ## Calculate Real quantities.
        #
        P_AVAIL                     = _P_AVAILS[GEARS - 1, range(len(V))]
        N_NORM                      = _N_NORMS[GEARS - 1, range(len(V))]
        RPM                         = _N_GEARS[GEARS - 1, range(len(V))]
        V_REAL                      = RPM / _GEAR_RATIOS[GEARS - 1, range(len(V))]
        RPM[RPM < n_idle] = n_idle

        results['gears']        = GEARS
        results['v_real']       = V_REAL
        results['p_available']  = P_AVAIL
        results['p_required']   = P_REQ
        results['rpm']          = RPM
        results['rpm_norm']     = N_NORM
        results['driveability'] = driveability_issues

        return model


#######################
##       MODEL       ##
#######################


    @property
    def model(self):
        return self._model

    def _set_model(self, mdl, skip_validation=False, validate_wltc_data=False):
        from wltp.model import _get_model_base, merge

        merged_model = _get_model_base()
        merge(merged_model, mdl)
        if not skip_validation:
            model.validate_model(merged_model, validate_wltc_data=validate_wltc_data)
        self._model = merged_model


    def driveability_report(self):
        cycle = self._model.get('cycle_run')
        if (not cycle is None):
            issues = []
            drv = cycle['driveability']
            pt = -1
            for t in drv.nonzero()[0]:
                if (pt+1 < t):
                    issues += ['...']
                issues += ['{:>4}: {}'.format(t, drv[t])]
                pt = t

            return '\n'.join(issues)
        return None





#######################
## PURE CALCULATIONS ##
##  Separate for     ##
##     testability!  ##
#######################


#    resistance_coeffs_regression_curves
def calc_default_resistance_coeffs(test_mass, regression_curves):
    a = regression_curves
    f0 = a[0][0] * test_mass + a[0][1]
    f1 = a[1][0] * test_mass + a[1][1]
    f2 = a[2][0] * test_mass + a[2][1]

    return (f0, f1, f2)

def addDriveabilityMessage(time_step, msg, driveability_issues):
    old = driveability_issues[time_step]
    driveability_issues[time_step] = old + msg


def addDriveabilityProblems(_GEARS_BAD, reason, driveability_issues):
    failed_steps = _GEARS_BAD.nonzero()[0]
    if (failed_steps.size != 0):
        log.info('%i %s: %s', failed_steps.size, reason, failed_steps)
        for step in failed_steps:
            addDriveabilityMessage(step, reason, driveability_issues)


def decideClass(wltc_data, p_m_ratio, v_max):
    '''

    @see: Annex 1, p 19
    '''
    class_limits            = {cl: (cd['pmr_limits'], cd.get('velocity_limits')) for (cl, cd) in wltc_data['classes'].items()}

    for (cls, ((pmr_low, pmr_high), v_limits)) in class_limits.items():
        if pmr_low < p_m_ratio <= pmr_high and \
                (not v_limits or v_limits[0] <= v_max < v_limits[1]):
            wltc_class      = cls
            break
    else:
        raise ValueError("Cannot determine wltp-class for PMR(%s)!\n  Class-limits(%s)" %(p_m_ratio, class_limits))

    return wltc_class



def calcDownscaleFactor(P_REQ, p_max_values, downsc_coeffs, dsc_v_split, p_rated, v_max, f_downscale_threshold):
    '''Check if downscaling required, and apply it.

    :return: (float) the factor

    @see: Annex 1-7, p 68
    '''

    ## Max required power
    #
    p_req_max               = P_REQ[p_max_values[0]]
    r_max                   = (p_req_max / p_rated)

    ## Cycle r0
    #
    if dsc_v_split is not None:
        assert              len(downsc_coeffs) == 2, downsc_coeffs

        downsc_coeffs       = downsc_coeffs[0] if (v_max <= dsc_v_split) else downsc_coeffs[1]

    if (downsc_coeffs is None):
        f_downscale         = 0
    else:
        (r0, a1, b1) = downsc_coeffs

        if (r_max < r0):
            f_downscale     = 0
        else:
            f_downscale     = a1 * r_max + b1
            f_downscale     = round(f_downscale, 1)
            if (f_downscale <= f_downscale_threshold):
                f_downscale = 0


    return f_downscale


def downscaleCycle(V, f_downscale, phases):
    '''Downscale just by scaling the 2 phases demarked by the 3 time-points with different factors,
    no recursion as implied by the specs.

    @see: Annex 1-7, p 64-68
    '''
    (t0, t1, t2)    = phases

    ## Accelaration phase
    #
    ix_acc          = np.arange(t0, t1 + 1)
    offset_acc      = V[t0]
    acc_scaled      = (1 - f_downscale) * (V[ix_acc] - offset_acc) + offset_acc
    assert          acc_scaled[0] == V[t0], ('smooth-start: ', acc_scaled[0],  V[t0])

    ## Decelaration phase
    #
    ix_dec          = np.arange(t1 + 1, t2 + 1)
    offset_dec      = V[t2]
    f_corr          = (acc_scaled[-1] - offset_dec) / (V[t1] - offset_dec)
    dec_scaled      = f_corr * (V[ix_dec] - offset_dec) + offset_dec
    assert          dec_scaled[-1] == V[t2], ('smooth-finish: ', dec_scaled[-1],  V[t2])

    scaled          = np.hstack((acc_scaled, dec_scaled))
    assert          (1 - f_downscale) * abs(scaled[t1 - t0] - scaled[t1 - t0 + 1]) <= abs(V[t1] - V[t1 + 1]), \
                                    ('smooth-tip: ', scaled[t1 - t0], scaled[t1 - t0 + 1], V[t1], V[t1 + 1])

    V_DSC           = np.hstack((V[:t0], scaled, V[t2 + 1:]))
    assert          V.shape == V_DSC.shape, _shapes(V, V_DSC)

    return V_DSC


def calcEngineRevs_required(V, gear_ratios, n_idle, v_stopped_threshold):
    '''Calculates the required engine-revolutions to achieve target-velocity for all gears.

    :return: array: _N_GEARS:   a (#gears X #velocity) float-array, eg. [3, 150] --> gear(3), time(150)
    :rtype: array: _GEARS:     a (#gears X #velocity) int-array, eg. [3, 150] --> gear(3), time(150)

    @see: Annex 2-3.2, p 71
    '''

    assert              V.ndim == 1, (V.shape, gear_ratios, n_idle)

    nG                  = len(gear_ratios)
    nV                  = len(V)

    _GEARS               = np.tile(np.arange(0, nG, dtype='int8')+ 1, (nV, 1)).T
    assert              _GEARS.shape == (nG, nV), (_GEARS.shape, gear_ratios, nV)

    _GEAR_RATIOS         = np.tile(gear_ratios, (nV, 1)).T       ## TODO: Prepend row for idle-gear in _N_GEARS
    _N_GEARS             = np.tile(V, (nG, 1)) * _GEAR_RATIOS
    assert              _GEARS.shape  == _GEAR_RATIOS.shape == _N_GEARS.shape, _shapes(_GEARS, _GEAR_RATIOS, _N_GEARS, V)

    stopped_steps                       = (V == 0)              ## (Annex 2-3.2 & Annex 2-4(a), p72)
    _N_GEARS [:, stopped_steps]          = n_idle
    _GEARS   [:, stopped_steps]          = 0

    return (_N_GEARS, _GEARS, _GEAR_RATIOS)


def possibleGears_byEngineRevs(V, A, _N_GEARS,
                               ngears, n_idle,
                               n_min_drive, n_min_gear2, n_max,
                               v_stopped_threshold,
                               driveability_issues):
    '''
    Calculates the engine-revolutions limits for all gears and returns for which they are accepted.

    My interpratation for Gear2 ``n_min`` limit::

                              _____________                ______________
                          ///INVALID///|   CLUTCHED   |  GEAR-2-OK
        EngineRevs(N): 0-----------------------+---------------------------->
        for Gear-2                     |       |      +--> n_clutch_gear2   := n_idle + MAX(
                                       |       |                                      0.15% * n_idle,
                                       |       |                                      3%    * n_range)
                                       |       +---------> n_idle
                                       +-----------------> n_min_gear2      := 90% * n_idle


    :return: _GEARS_YES: possibibilty for all the gears on each cycle-step
            (eg: [0, 10] == True --> gear(1) is possible for t=10)
    :rtype: list(booleans, nGears x CycleSteps)

    @see: Annex 2-3.2, p 71
    '''

    ## Identify impossible-gears by n_MAX.
    #
    GEARS_YES_MAX                           = (_N_GEARS <= n_max)
    GEARS_YES_MAX[-1, :]                    = True                          ## Exclude last gear from max-rule.
    _GEARS_BAD                              = (~GEARS_YES_MAX).all(axis=0)
    addDriveabilityProblems(_GEARS_BAD, 'g%i: Revolutions too high!' % ngears, driveability_issues)

    ## Replace impossibles with max-gear.
    #
    GEARS_YES_MAX[ngears - 1, _GEARS_BAD]    = True

    ## Identify impossible-gears by n_MIN.
    #
    ## TODO: Construct a matrix of n_min_drive for all gears, including exceptions for gears 1 & 2.
    #
    GEARS_YES_MIN           = (_N_GEARS >= n_min_drive)
    GEARS_YES_MIN[0, :]     = (_N_GEARS[0, :] >= n_idle) | (V <= v_stopped_threshold) # FIXME: move V==0 into own gear.

    ## NOTE: "interpratation" of specs for Gear-2
    #        and FIXME: NOVATIVE rule: "Clutching gear-2 only when Decelerating.".
    N_GEARS2                = _N_GEARS[1, :]
    GEARS_YES_MIN[1, :]     = ((N_GEARS2 >= n_min_gear2) & (A <= 0)) | ((N_GEARS2 >= n_min_gear2) & (A > 0))

    ## Revert impossibles to min-gear, n_min & clutched.
    #
    _GEARS_BAD = CLUTCH      = (~GEARS_YES_MIN).all(axis=0) ## TODO: Clutch on gear2 (f=will be fixed when add also CLUTCH_GEARS)
    GEARS_YES_MIN[0, _GEARS_BAD]                     = True # Revert to min-gear.
    #addDriveabilityProblems(_GEARS_BAD, 'g1: Revolutions too low!', driveability_issues)

    _GEARS_YES               = (GEARS_YES_MIN & GEARS_YES_MAX)

    return (_GEARS_YES, CLUTCH)


def calcPower_required(V, A, SLOPE, test_mass, f0, f1, f2, f_inertial):
    '''

    @see: Annex 2-3.1, p 71
    '''

    gee = 9.81

    VV      = V * V
    VVV     = VV * V
    assert  V.shape == VV.shape == VVV.shape == A.shape, _shapes(V, VV, VVV, A)

    if SLOPE is None:
        P_REQ   = (
            f0 * V + f1 * VV + f2 * VVV +
            f_inertial * A * V * test_mass
        ) / 3600.0
    else:
        assert  V.shape == SLOPE.shape, _shapes(V, SLOPE)
        P_REQ   = (
            f0 * V * np.cos(SLOPE) + f1 * VV + f2 * VVV +
            f_inertial * A * V * test_mass +
            test_mass * np.sin(SLOPE) * gee * V
        ) / 3600.0
    assert  V.shape == P_REQ.shape, _shapes(V, P_REQ)

    return P_REQ



def calcPower_available(_N_GEARS, n_idle, n_rated, p_rated, load_curve, p_safety_margin):
    '''

    @see: Annex 2-3.2, p 72
    '''

    _N_NORMS         = (_N_GEARS - n_idle) / (n_rated - n_idle)
    _P_WOTS          = np.interp(_N_NORMS, load_curve['n_norm'], load_curve['p_norm']) # When outside of load_curve, accept max-min gear.
#     from scipy.interpolate import interp1d
#     intrerp_f       = interp1d(load_curve[0], load_curve[1], kind='linear', bounds_error=False, fill_value=0, copy=False)
#     P_WOT           = intrerp_f(_N_NORMS)
    safety_margins   = np.tile(p_safety_margin, (_N_GEARS.shape[1], 1)).T
    _P_AVAILS        = _P_WOTS * safety_margins * p_rated

    return (_P_AVAILS, _N_NORMS)

def possibleGears_byPower(_N_GEARS, P_REQ,
                          n_idle, n_rated, p_rated, load_curve, p_safety_margin,
                          driveability_issues):
    '''

    @see: Annex 2-3.1 & 3.3, p 71 & 72
    '''

    (_P_AVAILS, _N_NORMS)      = calcPower_available(_N_GEARS, n_idle, n_rated, p_rated, load_curve, p_safety_margin)
    assert      _N_GEARS.shape == _P_AVAILS.shape, \
                                _shapes(P_REQ, _N_GEARS, _P_AVAILS)

    _GEARS_YES   = _P_AVAILS >= P_REQ

    _GEARS_BAD = (~_GEARS_YES).all(axis=0)
    addDriveabilityProblems(_GEARS_BAD, 'Insufficient power!', driveability_issues)

    return (_GEARS_YES, _P_AVAILS, _N_NORMS)

def selectGears(_GEARS, _G_BY_N, _G_BY_P, driveability_issues):
    assert                  _G_BY_N.shape == _G_BY_P.shape, _shapes(_G_BY_N, _G_BY_P)
    assert                  _G_BY_N.dtype == _G_BY_P.dtype == 'bool', _dtypes(_G_BY_N, _G_BY_P)

    _GEARS_YES               = _G_BY_N & _G_BY_P
    _GEARS[~_GEARS_YES]    = -1                ## NOTE: Invalid gears will be smoothed-away in extra_rule(1).
    assert                  _G_BY_N.dtype == _G_BY_P.dtype == 'bool', _dtypes(_G_BY_N, _G_BY_P)
    GEARS                   = _GEARS.max(axis=0)

    return GEARS




_escape_char = 128

_regex_gears2regex = re.compile(br'\\g(\d+)')

PY2 = sys.version_info[0] < 3
if PY2:
    def dec_byte_repl(m):
        return chr(_escape_char + int(m.group(1)))
else:
    def dec_byte_repl(m):
        return bytes([_escape_char + int(m.group(1))])


def gearsregex(gearspattern):
    '''
    :param gearspattern: regular-expression or substitution that escapes decimal-bytes written as: ``\g\d+``
                        with adding +128, eg::

                            \g124|\g7 --> unicode(128+124=252)|unicode(128+7=135)
    '''

    assert          isinstance(gearspattern, bytes), 'Not bytes: %s' % gearspattern # For python-2 to work with __future__.unicode_literals.

    regex           = _regex_gears2regex.sub(dec_byte_repl, gearspattern)
    return          re.compile(regex)


def np2bytes(NUMS):
    if (NUMS < 0).any() or (NUMS >= (256 - _escape_char)).any():
        assert          all(NUMS >= 0)  and all(NUMS < (256 - _escape_char)), 'Outside byte-range: %s' % NUMS[(NUMS < 0) | (NUMS >= (256 - _escape_char))]

    return          (NUMS + _escape_char).astype('uint8').tostring()


def bytes2np(bytesarr):
    assert          isinstance(bytesarr, bytes), 'Not bytes: %s' % bytesarr

    return          np.fromstring(bytesarr, dtype='uint8') - _escape_char


def assert_regexp_unmatched(regex, string, msg):
    assert not re.findall(regex, string), \
            '%s: %s' % (msg, [(m.start(), m.group()) for m in re.finditer(regex, string)])


#=====================
# Driveability rules #
#=====================

def rule_checkSingletons(bV, GEARS, CLUTCH, driveability_issues, re_zeros):
    re_singletons = gearsregex(b'(\g0)')

def rule_a(bV, GEARS, CLUTCH, driveability_issues, re_zeros):
    """Rule (a): Clutch & set to 1st-gear before accelerating from standstill.

     Implemented with a regex, outside rules-loop:
     Also ensures gear-0 always followed by gear-1.

     NOTE: Rule(A) not inside x2 loop, and last to run.
    """

    for m in re_zeros.finditer(bV):
        t_accel                 = m.end()
        # Exclude zeros at the end.
        if (t_accel == len(bV)):
            break
        GEARS[t_accel - 1:t_accel]          = 1
        CLUTCH[t_accel - 1:t_accel - 2]     = True
        addDriveabilityMessage(t_accel-1, '(a: X-->0)', driveability_issues)
    assert_regexp_unmatched(b'\x00[^\x00\x01]', GEARS.astype('uint8').tostring(), 'Jumped gears from standstill')


def step_rule_b1(t, pg, g, V, A, GEARS, driveability_issues):
    """Rule (b1): Do not skip gears while accelerating."""

    if ((pg+1) < g and A[t-1] > 0):
        pg          = pg+1
        GEARS[t]    = pg
        addDriveabilityMessage(t, '(b1: %i-->%i)' % (g, pg), driveability_issues)
        return True
    return False


def step_rule_b2(t, pg, g, V, A, GEARS, driveability_issues):
    """Rule (b2): Hold gears for at least 2sec when accelerating."""

    if pg != GEARS[t-2]:  # A[t-1] > 0):  NOTE: Not checking Accel on the final step A[t-1]!
        #assert g > pg, 'Rule e & g missed downshift(%i: %i-->%i) in acceleration!'%(t, pg, g)
        if (g < pg):
            addDriveabilityMessage(t, 'Rule e or g missed downshift(%i: %i-->%i) in acceleration?' %
                                   (t, pg, g), driveability_issues)
        if A[t-2] > 0:
            addDriveabilityMessage(t, '(b2: %i-->%i)' % (g, pg), driveability_issues)

            return True
    return False


def step_rule_c1(t, pg, g, V, A, GEARS, driveability_issues):
    """Rule (c1): Skip gears <3sec when decelerating. """

    if ((pg != GEARS[t-3:t-1]).any() and (A[t-2:t] < 0).all()):
        pt = t - 2
        while (pt >= t-3 and GEARS[pt] == pg):
            pt -= 1

        ## Skip even further...
        #
        if (GEARS[t+1] < g):
            t += 1
            g = GEARS[t]

        GEARS[pt+1:t] = g
        for tt in range(pt+1, t):
            addDriveabilityMessage(tt, '(c1: %i-->%i)' % (pg, g), driveability_issues)
        return True
    return False


def rule_c2(bV, A, GEARS, CLUTCH, driveability_issues, re_zeros):
    """Rule (c2): Skip 1st-gear while decelerating to standstill.

     Implemented with a regex, outside rules-loop:
     Search for zeros in _reversed_ V & GEAR profiles,
     for as long Accel is negative.
     NOTE: Rule(c2) is the last rule to run.
    """

    nV          = len(bV)
    for m in re_zeros.finditer(bV[::-1]):
        t_stop  = m.end()
        # Exclude zeros at the end.
        if (t_stop == nV):
            break
        t = nV - t_stop - 1
        while (A[t] < 0 and GEARS[t] == 1):
            addDriveabilityMessage(t, '(c2: %i-->0)'% GEARS[t], driveability_issues)
            GEARS[t]   = 0
            CLUTCH[t]  = False
            t -= 1


def step_rule_d(t, pg, g, V, A, GEARS, driveability_issues):
    """Rule (d): Cancel shifts after peak velocity."""

    if (A[t-2] > 0 and  A[t-1] < 0 and GEARS[t-2] == pg):
        GEARS[t]    = pg
        addDriveabilityMessage(t, '(d: %i-->%i)' % (g, pg), driveability_issues)
        return True
    return False


def step_rule_e(t, pg, g, V, A, GEARS, driveability_issues):
    """Rule (e): Cancel shifts lasting 5secs or less."""

    if (pg > g):
        ## Travel back in time for 5secs.
        #
        pt = t-2
        while (pt >= t-5 and GEARS[pt] == pg):
            pt -= 1

        if (GEARS[pt] < pg): # NOTE: Apply rule(e) also for any LOWER initial/final gears (not just for i-1).
            GEARS[pt+1:t] = g
            for tt in range(pt+1, t):
                addDriveabilityMessage(tt, '(e: %i-->%i)' % (pg, g), driveability_issues)
            return True
    return False


def step_rule_f(t, pg, g, V, A, GEARS, driveability_issues):
    """Rule(f): Cancel 1sec downshifts (under certain circumstances)."""

    if (pg < g and GEARS[t-2] >= g):
        # NOTE: Nowhere to apply it since rule(b2) would have eliminated 1-sec shifts.  Moved before rule(b)!
        # NOTE: Applying rule(f) also for i-2, i-3, ... signular-downshifts.
        # FIXME: Rule(f) implement further constraints.
        # NOTE: Rule(f): What if extra conditions unsatisfied? Allow shifting for 1 sec only??
        GEARS[t-1] = min(g, GEARS[t-2])
        addDriveabilityMessage(t-1, '(f: %i-->%i)' % (pg, g), driveability_issues)
        return True

    return False


def step_rule_g(t, pg, g, V, A, GEARS, driveability_issues):
    """Rule(g): Cancel upshift during acceleration if later downshifted for at least 2sec."""

    if (pg > g and (GEARS[t:t+2] == g).all() and (A[t-1:t+2] > 0).all()):
        ## Travel back in time for as long accelerating and same gear.
        #
        pt = t-2
        while (GEARS[pt] == pg and A[pt] > 0):
            pt -= 1

        GEARS[pt+1:t] = g
        for tt in range(pt+1, t):
            addDriveabilityMessage(tt, '(g: %i-->%i)' % (pg, g), driveability_issues)
        return True

    return False


def applyDriveabilityRules(V, A, GEARS, CLUTCH, driveability_issues):
    '''
    @note: Modifies GEARS & CLUTCH.
    @see: Annex 2-4, p 72
    '''

    def apply_step_rules(rules, isStopOnFirstApplied):
        for t in t_range:
            if (GEARS[t-1] != GEARS[t]):       ## All rules triggered by a gear-shift.
                for rule in rules:
                        if (rule(t, GEARS[t-1], GEARS[t], V, A, GEARS, driveability_issues) and \
                                            (isStopOnFirstApplied or GEARS[t-1] == GEARS[t])):
                            break

    ## V --> byte-array to search by regex.
    #
    V                           = V.copy(); V[V > (255 - _escape_char)] = (255 - _escape_char)
    bV                          = np2bytes(V)
    re_zeros                    = gearsregex(br'\g0+')

    ## NOTE: Extra_rule(1): Smooth-away INVALID-GEARS.
    #
    for t in range(2, len(GEARS)):      # Start from 2nd element to accomodate rule(e)'s backtracking.
        if (GEARS[t] < 0):
            GEARS[t] = GEARS[t-1]

    ## Loop X 2 driveability-rules.
    #
    t_range = range(5, len(GEARS))      # Start from 5th element to accomodate rule(e)'s backtracking.
    for _ in [0, 1]:
        apply_step_rules([step_rule_g, step_rule_f], False)                   # NOTE: Rule-order and first-to-apply flag unimportant.
        apply_step_rules([step_rule_e, step_rule_b1, step_rule_b2], False)    # NOTE: Rule-order for b1 &b2 unimportant.
        apply_step_rules([step_rule_c1], False)

        rule_c2(bV, A, GEARS, CLUTCH, driveability_issues, re_zeros)

    rule_a(bV, GEARS, CLUTCH, driveability_issues, re_zeros)


def run_cycle(V, A, P_REQ, gear_ratios,
                   n_idle, n_min_drive, n_rated,
                   p_rated, load_curve,
                   params):
    '''Calculates gears, clutch and actual-velocity for the cycle (V).
    Initial calculations happen on engine_revs for all gears, for all time-steps of the cycle (_N_GEARS array).
    Driveability-rules are applied afterwards on the selected gear-sequence, for all steps.

    :param V: the cycle, the velocity profile
    :param A: acceleration of the cycle (diff over V) in m/sec^2
    :return: CLUTCH:    a (1 X #velocity) bool-array, eg. [3, 150] --> gear(3), time(150)

    :rtype: array
    '''

    ## A multimap to collect problems.
    #
    driveability_issues         = np.empty_like(V, dtype='object')
    driveability_issues[:]      = ''



    ## Read and calc model parameters.
    #
    n_range                     = (n_rated - n_idle)

    f_n_max                     = params.get('f_n_max', 1.2)
    n_max                       = n_idle + f_n_max * n_range

    if n_min_drive is None:
        f_n_min                 = params.get('f_n_min', 0.125)
        n_min_drive             = n_idle + f_n_min * n_range

    f_n_min_gear2               = params.get('f_n_min_gear2', 0.9)
    n_min_gear2                 = f_n_min_gear2 * n_idle

    f_n_clutch_gear2            = params.get('f_n_clutch_gear2', [1.15, 0.03])
    n_clutch_gear2              = max(f_n_clutch_gear2[0] * n_idle, f_n_clutch_gear2[1] * n_range + n_idle)

    p_safety_margin             = params.get('f_safety_margin', 0.9)
    v_stopped_threshold         = params.get('v_stopped_threshold', 1) # Km/h


    (_N_GEARS, _GEARS, \
            _GEAR_RATIOS)       = calcEngineRevs_required(V, gear_ratios, n_idle, v_stopped_threshold)

    (_G_BY_N, CLUTCH)           = possibleGears_byEngineRevs(V, A, _N_GEARS,
                                                len(gear_ratios), n_idle,
                                                n_min_drive, n_min_gear2, n_max,
                                                v_stopped_threshold,
                                                driveability_issues)

    (_G_BY_P, _P_AVAILS, _N_NORMS) = possibleGears_byPower(_N_GEARS, P_REQ,
                                                n_idle, n_rated, p_rated, load_curve, p_safety_margin,
                                                driveability_issues)

    assert _GEAR_RATIOS.shape == _N_GEARS.shape == _P_AVAILS.shape == _N_NORMS.shape, \
                                _shapes(_GEAR_RATIOS, _N_GEARS, _P_AVAILS, _N_NORMS)


    GEARS                       = selectGears(_GEARS, _G_BY_N, _G_BY_P, driveability_issues)
    CLUTCH[(GEARS == 2) & (_N_GEARS[1, :] < n_clutch_gear2)] = True

    assert V.shape == GEARS.shape, _shapes(V, GEARS)
    assert GEARS.shape == CLUTCH.shape == driveability_issues.shape, \
                                _shapes(GEARS, CLUTCH.shape, driveability_issues)
    assert 'i' == GEARS.dtype.kind, GEARS.dtype
    assert ((GEARS >= -1) & (GEARS <= len(gear_ratios))).all(), (min(GEARS), max(GEARS))


    return GEARS, CLUTCH, _GEAR_RATIOS, _N_GEARS, _P_AVAILS, _N_NORMS, driveability_issues



if __name__ == '__main__':
    pass



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

from matplotlib import pyplot as plt
import os

from wltp import plots
from wltp.test import wltp_db_tests as wltpdb


def data_meanN_pmr(gened_column, heinz_column):
    gened_column='rpm'
    heinz_column='n'
    vehdata = wltpdb.aggregate_single_columns_means(gened_column, heinz_column)

    return vehdata


if __name__ == '__main__':
    os.chdir(os.path.join(wltpdb.mydir, wltpdb.samples_dir))

    classes = [
        ['class1',  dict(data_fmt='4k', data_kws=dict(fillstyle='none'), ref_fmt='4g', ref_kws=dict(fillstyle='none'))],  
        ['class2',  dict(data_fmt='_k', data_kws=dict(fillstyle='none'), ref_fmt='_g', ref_kws=dict(fillstyle='none'))],  
        ['class3a', dict(data_fmt='ok', data_kws=dict(fillstyle='none'), ref_fmt='og', ref_kws=dict(fillstyle='none'))], 
        ['class3b', dict(data_fmt='+k', data_kws=dict(fillstyle='none'), ref_fmt='+g', ref_kws=dict(fillstyle='none'), diff_label='Differences')], 
    ]
    
    vehdata = data_meanN_pmr(gened_column='rpm', heinz_column='n')
    
    axes_tuple = None
    for (cls, kws) in classes:
        cls_data = vehdata[vehdata['wltc_class'] == cls]
        if cls_data.empty:
            continue
        (X, Y, Y_REF) = cls_data.pmr, cls_data.gened, cls_data.heinz
    
        axes_tuple, artists = plots.plot_xy_diffs_scatter(
            X, Y, X, Y_REF,
            ref_label='Access-db %s'%cls, data_label='Python %s'%cls,
            title="Python vs Access-db(2sec rule)",
            x_label=r'$PMR [W/kg]$',
            y_label='$EngineSpeed [rpm]$',
            axes_tuple=axes_tuple,
            **kws
        )

    axes_tuple[1].legend(loc=4, fancybox=True, framealpha=0.5)
    axes_tuple[0].legend(loc=1, fancybox=True, framealpha=0.5)

    plt.show()

